

import ..YM.flw, ..YM.force_gauge



function flw(U, psi, int::FlowIntr{NI,T}, ns::Int64, eps, gp::GaugeParm, dpar::DiracParam, lp::SpaceParm, ymws::YMworkspace, dws::DiracWorkspace) where {NI,T}    
    @timeit "Integrating flow equations" begin
        for i in 1:ns
            force_gauge(ymws, U, int.c0, 1, gp, lp)

            if int.add_zth
                add_zth_term(ymws::YMworkspace, U, lp)
            end

            ymws.mom .= ymws.frc1
            U .= expm.(U, ymws.mom, 2*eps*int.r)

            Nablanabla!(dws.sAp, U, psi, dpar, dws, lp)
            dws.sp .= dws.sAp 
            psi .= psi + 2*int.r*eps*dws.sp

            for k in 1:NI
                force_gauge(ymws, U, int.c0, 1, gp, lp)

                if int.add_zth
                    add_zth_term(ymws::YMworkspace, U, lp)
                end

                ymws.mom .= int.e0[k].*ymws.mom .+ int.e1[k].*ymws.frc1
                U .= expm.(U, ymws.mom, 2*eps)

                Nablanabla!(dws.sp, U, psi, dpar, dws, lp)

                dws.sAp .= int.e0[k].*dws.sAp .+ int.e1[k].*dws.sp
                psi .= psi + 2*eps*dws.sAp
            end
        end
    end
    
    return nothing
end
flw(U, psi, int::FlowIntr{NI,T}, ns::Int64, gp::GaugeParm, dpar::DiracParam, lp::SpaceParm, ymws::YMworkspace, dws::DiracWorkspace) where {NI,T} = flw(U, psi, int::FlowIntr{NI,T}, ns::Int64, int.eps, gp::GaugeParm, dpar::DiracParam, lp::SpaceParm, ymws::YMworkspace, dws::DiracWorkspace)



function Nablanabla!(so, U, si, dpar::DiracParam, dws::DiracWorkspace, lp::SpaceParm{4,6,B,D}) where {B,D}  # TO BE CONTINUED

    if B == BC_SF_AFWB || B == BC_SF_ORBI
        @timeit "Nablanabla" begin
        
            @timeit "SF fix" begin
            CUDA.@sync begin
                CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_sfbndfix!(si,si,1.0,Gamma{16},lp)
            end
            end

            @timeit "Fermion force flow" begin
            CUDA.@sync begin
                CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_Nablanabla!(so, U, si, dpar.th, lp)
            end
            end

            @timeit "SF fix" begin
            CUDA.@sync begin
                CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_sfbndfix!(so,si,dpar.ct,Gamma{5},lp)
            end
            end
        end


    else
        @timeit "DwdagDw" begin
            CUDA.@sync begin
                CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_Nablanabla!(so, U, si, dpar.th, lp)
            end

        end
    end 
    
    return nothing
end


function krnl_Nablanabla!(so, U, si, th, lp::SpaceParm{4,6,B,D}) where {B,D}

    b = Int64(CUDA.threadIdx().x);  r = Int64(CUDA.blockIdx().x)

    @inbounds begin 

        so[b,r] = -8*si[b,r]

	        bu1, ru1 = up((b,r), 1, lp)
            bd1, rd1 = dw((b,r), 1, lp)
            bu2, ru2 = up((b,r), 2, lp)
            bd2, rd2 = dw((b,r), 2, lp)
            bu3, ru3 = up((b,r), 3, lp)
            bd3, rd3 = dw((b,r), 3, lp)
            bu4, ru4 = up((b,r), 4, lp)
            bd4, rd4 = dw((b,r), 4, lp)


        so[b,r] += (th[1] * (U[b,1,r]*si[bu1,ru1]) +conj(th[1]) * (U[bd1,1,rd1]*si[bd1,rd1]) +
                    th[2] * (U[b,2,r]*si[bu2,ru2]) +conj(th[2]) * (U[bd2,2,rd2]*si[bd2,rd2]) +
                    th[3] * (U[b,3,r]*si[bu3,ru3]) +conj(th[3]) * (U[bd3,3,rd3]*si[bd3,rd3]) +
                    th[4] * (U[b,4,r]*si[bu4,ru4]) +conj(th[4]) * (U[bd4,4,rd4]*si[bd4,rd4])  )

    end
        
    return nothing
end

function Dslash_sq!(so, U, si, dpar::DiracParam, dws::DiracWorkspace, lp::SpaceParm{4,6,B,D}) where {B,D}

    if B == BC_SF_AFWB || B == BC_SF_ORBI
        @timeit "DwdagDw" begin
        
            @timeit "SF fix" begin
            CUDA.@sync begin
                CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_sfbndfix!(si,si,1.0,Gamma{16},lp)
            end
            end

            @timeit "g5Dslsh" begin
            CUDA.@sync begin
                CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_g5Dslsh!(dws.st, U, si, dpar.th, lp)
            end
            end

            @timeit "SF fix" begin
            CUDA.@sync begin
                CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_sfbndfix!(dws.st,si,dpar.ct,Gamma{5},lp)
            end
            end

            if abs(dpar.csw) > 1.0E-10
                @timeit "Dw_improvement" begin
                        CUDA.@sync begin
                            CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_g5Dslsh_impr!(dws.st, dws.csw, dpar.csw, si,  lp)
                        end
                end
            end


            @timeit "g5Dslsh" begin
            CUDA.@sync begin
                CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_g5Dslsh!(so, U, dws.st, dpar.th, lp)
            end
            end


            @timeit "SF fix" begin
            CUDA.@sync begin
                CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_sfbndfix!(so,dws.st,dpar.ct,Gamma{5},lp)
            end
            end

            if abs(dpar.csw) > 1.0E-10
                @timeit "Dw_improvement" begin
                        CUDA.@sync begin
                            CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_g5Dslsh_impr!(so, dws.csw, dpar.csw, dws.st,  lp)
                        end
                end
            end               

            
        end

    else
        @timeit "DwdagDw" begin
            CUDA.@sync begin
                CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_g5Dslsh!(dws.st, U, si,  dpar.th, lp)
            end

            if abs(dpar.csw) > 1.0E-10
                @timeit "Dw_improvement" begin
                        CUDA.@sync begin
                            CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_g5Dslsh_impr!(dws.st, dws.csw, dpar.csw, si,  lp)
                        end
                end
            end               

            CUDA.@sync begin
                CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_g5Dslsh!(so, U, dws.st,  dpar.th, lp)
            end

            if abs(dpar.csw) > 1.0E-10
                @timeit "Dw_improvement" begin
                        CUDA.@sync begin
                            CUDA.@cuda threads=lp.bsz blocks=lp.rsz krnl_g5Dslsh_impr!(so, dws.csw, dpar.csw, dws.st,  lp)
                        end
                end
            end               

        end
    end 
    
    return nothing
end


function krnl_g5Dslsh!(so, U, si, th, lp::SpaceParm{4,6,B,D}) where {B,D}

    b = Int64(CUDA.threadIdx().x);  r = Int64(CUDA.blockIdx().x)

    @inbounds begin 

        so[b,r] = 4*si[b,r]

	        bu1, ru1 = up((b,r), 1, lp)
            bd1, rd1 = dw((b,r), 1, lp)
            bu2, ru2 = up((b,r), 2, lp)
            bd2, rd2 = dw((b,r), 2, lp)
            bu3, ru3 = up((b,r), 3, lp)
            bd3, rd3 = dw((b,r), 3, lp)
            bu4, ru4 = up((b,r), 4, lp)
            bd4, rd4 = dw((b,r), 4, lp)


        so[b,r] -= 0.5*( th[1] * U[b,1,r]*si[bu1,ru1] +conj(th[1]) * U[bd1,1,rd1]*si[bd1,rd1] +
                         th[2] * U[b,2,r]*si[bu2,ru2] +conj(th[2]) * U[bd2,2,rd2]*si[bd2,rd2] +
                         th[3] * U[b,3,r]*si[bu3,ru3] +conj(th[3]) * U[bd3,3,rd3]*si[bd3,rd3] +
                         th[4] * U[b,4,r]*si[bu4,ru4] +conj(th[4]) * U[bd4,4,rd4]*si[bd4,rd4]  )

        so[b,r] = dmul(Gamma{5}, so[b,r])
    end
        
    return nothing
end

function krnl_g5Dslsh_impr!(so, Fcsw, csw, si, lp::SpaceParm{4,6,B,D}) where {B,D}

    @inbounds begin 

    b = Int64(CUDA.threadIdx().x);  
    r = Int64(CUDA.blockIdx().x)

    I = point_coord((b,r), lp)

    TWP  = false

        if TWP
        so[b,r] += 0.5*csw*im*dmul(Gamma{5},( Fcsw[b,1,r]*dmul(Gamma{10},si[b,r]) + Fcsw[b,2,r]*dmul(Gamma{11},si[b,r]) + Fcsw[b,3,r]*dmul(Gamma{12},si[b,r])
                                          -Fcsw[b,4,r]*dmul(Gamma{15},si[b,r]) - Fcsw[b,5,r]*dmul(Gamma{14},si[b,r]) - Fcsw[b,6,r]*dmul(Gamma{13},si[b,r])))    
        else   
        so[b,r] += 0.5*csw*im*dmul(Gamma{5},( Fcsw[b,1,r]*dmul(Gamma{10},si[b,r]) + Fcsw[b,2,r]*dmul(Gamma{11},si[b,r]) + Fcsw[b,3,r]*dmul(Gamma{12},si[b,r])
                                          -Fcsw[b,4,r]*dmul(Gamma{15},si[b,r]) - Fcsw[b,5,r]*dmul(Gamma{14},si[b,r]) - Fcsw[b,6,r]*dmul(Gamma{13},si[b,r])))
        end
    end
        
    return nothing
end